
#ifndef _REGRESSION_H
#define	_REGRESSION_H

#include <iostream>
#include <iomanip>
#include <cmath>

#include "plink.h"
#include "helper.h"
#include "stats.h"
static const double vif_threshold = 50;


class LogisticRegression
{

    int nind; // Number of invididuals in this sample
    int np;  // Main effects + interaction + intercept

    vector_t coef;  // beta
    matrix_t S;     // Sigma

    vector<int> Y;
    vector<vector<double> > X;
    vector<string> label;

    vector_t p;
    vector_t V; // diagonal p(1-p)

    bool all_valid, converge;

    Locus *snp;

public:

    ///////////////////////////

    // Default additive coding
    static const int mAA = 0, mAB = 1, mBB = 2; // Coding of genotypes;

    LogisticRegression() { init(); }

    void init()
    {
    }

    void buildDesignMatrix(const vector<Individual*> &persons, Locus *snp)
    {
	this->snp = snp;

	Y.reserve(persons.size());
	X.reserve(persons.size());

	label.push_back("INT");
	label.push_back("ADD");

	int id = snp->id;

        // 
	np = 2 + persons[0]->covariates.size();

	vector <double> data(np, 0);

	for (int i = 0; i < persons.size(); i++) {
	    int geno = persons[i]->getGeno(id);

	    data[0] = 1; // Intercept

	    if(geno == G_MM) continue;

	    if(geno == G_AA)
		data[1] = (mAA);
	    else if(geno == G_aA)
		data[1] = (mAB);
	    else if(geno == G_aa)
		data[1] = (mBB);

            int nomissing = 1;

            for(int j=0;j < persons[i]->covariates.size(); j++)
                if(persons[i]->covariates[j] == NA) nomissing = 0;

            if(nomissing == 0) continue;

	    // Add covariates here
	    for(int j=0;j < persons[i]->covariates.size(); j++) {
		data[2 + j] = persons[i]->covariates[j];
	    }

	    X.push_back(data);

	    if(persons[i]->phenotype == 1)
		Y.push_back(0);
	    else
		Y.push_back(1);

	}

	nind = Y.size();

	p.resize(nind);
	V.resize(nind);

	all_valid = checkVIF();
	if(all_valid == false) {
	    cout << "Warning: VIF-based check for multicollinearity failed (snp " << snp->name << ").\n";
	}


	if (par::verbose) {
	    cout << "X design matrix\n";
	    display(X);
	    cout << "\n";
	}
    }

    void fitLM()
    {
      coef.resize(np);
      sizeMatrix(S,np,np);

      if (np==0 || nind==0 )
        return;

      if (par::verbose)
        {
          for (int i=0; i<nind; i++)
    	{
    	  cout << i << "\t"
    	       << Y[i] << "\t";
    	  for (int j=0; j<np; j++)
    	    cout << X[i][j] << "\t";
    	  cout << "\n";
    	}
        }

      ///////////////////////////////////////
      // Newton-Raphson to fit logistic model

        converge = false;
        int it = 0;

        while ( ! converge && it < 20 )
          {

    	// Determine p and V
    	for (int i=0; i<nind; i++)
    	  {
    	    double t = 0;
    	    for (int j=0; j<np; j++)
    	      t += coef[j] * X[i][j];
    	    p[i] = 1/(1+exp(-t));
    	    V[i] = p[i] * (1-p[i]);
    	  }

    	// Update coefficients
    	// b <- b +  solve( t(X) %*% V %*% X ) %*% t(X) %*% ( y - p )

    	matrix_t T;
    	sizeMatrix(T,np,np);

    	for (int j=0; j<np; j++)
    	  for (int k=j; k<np; k++)
    	    {
    	      double sum = 0;
    	      for (int i=0; i<nind; i++)
    		sum += X[i][j] * V[i] * X[i][k] ;
    	      T[j][k] = T[k][j] = sum;
    	    }

    	T = svd_inverse(T);

    	matrix_t T2;
    	// Resize and set elements to 0
    	sizeMatrix(T2,np,nind);

    	// note implicit transpose of X
    	for (int i=0; i<np; i++)
    	  for (int j=0; j<nind; j++)
    	    for (int k=0; k<np; k++)
    	      T2[i][j] += T[i][k] * X[j][k];

    	vector_t t3(nind);
    	for (int i=0; i<nind; i++)
    	  t3[i] = Y[i] - p[i];

    	vector_t ncoef(np);
    	for (int j=0; j<np; j++)
    	  for (int i=0; i<nind; i++)
    	    ncoef[j] += T2[j][i] * t3[i];

    	// Update coefficients, and check for
    	// convergence
    	double delta = 0;
    	for (int j=0; j<np; j++)
    	  {
    	    delta += abs(ncoef[j]);
    	    coef[j] += ncoef[j];
    	  }

    	if ( delta < 1e-6 )
    	  converge = true;

    	// Next iteration
    	it++;
          }

        if( it > 19 )
            cout << "Warning: Logistic regression didn't converge after " << it << " iterations for snp: " << snp->name << ". \n";

        if(it > 19) {

            double mean = 0, mean0 = 0, mean1 = 0, n0 = 0, n1 = 0;
            for (int i = 0; i < nind; i++) {
                mean += X[i][1];
                if(Y[i] == 0) { mean0 += X[i][1]; n0 ++; }
                if(Y[i] == 1) { mean1 += X[i][1]; n1 ++; }
            }

            // cout << "Means: " << (mean/nind) << " " << (mean0/n0) << " " << (mean1/n1) << "\n";
        }


        /////////////////////////////////////////
        // Obtain covariance matrix of estimates

        // S <- solve( t(X) %*% V %*% X )

        // Transpose X and multiple by diagonal V
        matrix_t Xt;
        sizeMatrix(Xt, np, nind);
        for (int i=0; i<nind; i++)
          for (int j=0; j<np; j++)
    	Xt[j][i] = X[i][j] * V[i];

        multMatrix(Xt,X,S);
        S = svd_inverse(S);

        if (par::verbose) {
            cout << "beta\n";
            display(coef);
            cout << "Sigma\n";
            display(S);
            cout << "\n";
        }
    }

    // VIF-based check for multicollinearity
    bool checkVIF()
    {

      // Calculate correlation matrix for X
      // Skip intercept

      int p = X.size();
      if (p<2) return false;

      int q = X[0].size() - 1;
      if ( q < 2 ) return true;

      vector_t m(q);
      matrix_t c;
      sizeMatrix(c,q,q);

      for (int i=0; i<p; i++)
        for (int j=0; j<q; j++)
          m[j] += X[i][j+1];

      for (int j=0; j<q; j++)
        m[j] /= (double)p;

      for (int i=0; i<p; i++)
        for (int j1=0; j1<q; j1++)
          for (int j2=j1; j2<q; j2++)
    	c[j1][j2] += ( X[i][j1+1] - m[j1] ) * ( X[i][j2+1] - m[j2] );

      for (int j1=0; j1<q; j1++)
        for (int j2=j1; j2<q; j2++)
          c[j1][j2] /= (double)(p-1);

      for (int j1=0; j1<q; j1++)
        for (int j2=j1+1; j2<q; j2++)
          {
    	c[j1][j2] /= sqrt( c[j1][j1] * c[j2][j2] );
    	c[j2][j1] = c[j1][j2];

     	if ( c[j2][j1] > 0.999 )
     	  {
    	    if (par::verbose)
    	      cout << "individual element > 0.999, skipping VIF\n";
    	    return false;
    	  }
          }


      // Any item with zero variance?

      for (int j=0; j<q; j++)
        {
          if ( c[j][j] == 0 || ! realnum( c[j][j] ) )
    	return false;
          c[j][j] = 1;
        }

      // Get inverse
      c = svd_inverse(c);

      if (par::verbose)
        {
          cout << "VIF on diagonals\n";
          display(c);
          cout << "\n";
        }

      // Calculate VIFs
      double maxVIF = 0;
      for (int j=0;j<q;j++)
        {

          // r^2 = 1 - 1/x where x is diagonal element of inverted
          // correlation matrix
          // As VIF = 1 / ( 1 - r^2 ) , implies VIF = x

          if ( c[j][j] > vif_threshold )
    	return false;
        }


      return true;

    }

    vector_t getVar()
    {
      vector_t var(np);
      for (int i=0; i<np; i++)
        var[i] = S[i][i];
      return var;
    }

    bool checkCollinearity()
    {
	bool result = validParameters();

	if(result == false) {
	    cout << "Warning: Multi-Collinearity in parameters while performing regression for snp "
		<< snp->name << "\n";
	}

	return result;
    }

    bool validParameters()
    {

      // Empty model?
      if (np==0 || nind==0)
        {
          all_valid = false;
          return all_valid;
        }


      // Display covariance matrix in verbose mode

      if ( par::verbose )
        {
          cout << "Covariance matrix of estimates\n";
          display(S);
          cout << "\n";
        }

      // Check for multicollinearity

      // For each term, see that estimate is not too strongly (r>0.99)
      // correlated with another, starting at last


      for (int i = 1; i<np; i++)
        {
          if ( S[i][i] < 1e-20 ) { all_valid = false; }
          else if ( ! realnum(S[i][i]) ) { all_valid = false; }
        }

      if ( all_valid )
        for (int i = np-1; i>0; i--)
          {
    	for (int j = i-1; j>=0; j--)
    	  {
    	    if ( S[i][j] / sqrt( S[i][i] * S[j][j] ) > 0.99999 )
    	      {
    		cout << "Error: Collinearity between " << i << " and " << j << "\n";
    		all_valid = false;
    		break;
    	      }
    	  }
          }
      return all_valid;
    }

    LogisticOutput getResults()
    {
	LogisticOutput L;

	L.snp = snp;
	L.a1 = snp->allele1;

	int p = 1;
	double var = S[p][p];

	bool okay = var < 1e-20 || !realnum(var) ? false : all_valid;

        if (!converge) {
            cout << "Warning: Could not get odds ratio for SNP " << snp->name << ". Setting OR to 1.\n";
            L.odds = 1;
            L.p = 1;
            return L;
        }
        

	if (!okay) {
	    cout << "Var = " << S[p][p] << " for snp " << snp->name << "\n";
	    cout << "Warning: Could not get odds ratio for SNP " << snp->name << "\n";
	    L.odds = NA;
	    L.p = NA;
	    return L;
	}

	double se = 0;
	double Z = 0;
	double pvalue = 1;

	if (okay) {
	    se = sqrt(var);
	    Z = coef[p] / se;
	    //	  pvalue = pT(Z,Y.size()-np);
	    pvalue = chiprobP(Z * Z, 1);
	}

	L.odds = exp(coef[p]);
	L.p = pvalue;

	// cout << "Var = " << S[p][p] << " for snp " << snp->name << " P=" << pvalue << " OR=" << exp(coef[p]) << "\n";

	return L;
    }

    void displayResults(ostream & OUT, const Locus * loc)
    {

      vector_t var;

      if ( all_valid )
        var = getVar();
      else
        {
          var.clear();
          var.resize(np,0);
        }


      for (int p=1; p<np; p++) // Skip intercept
        {

          bool okay = var[p] < 1e-20 || !realnum(var[p]) ? false : all_valid;

          double se = 0;
          double Z = 0;
          double pvalue = 1;

          if (okay)
    	{
    	  se = sqrt(var[p]);
    	  Z = coef[p] / se;
    	  //	  pvalue = pT(Z,Y.size()-np);
    	  pvalue = chiprobP(Z*Z,1);
    	}

          // If filtering p-values
          //if ( (!par::pfilter) || pvalue <= par::pfvalue )
    	{

    	  OUT << setw(4) << loc->chr  << " "
    	      << setw(par::pp_maxsnp) << loc->name << " "
    	      << setw(10) << loc->bp << " "
    	      << setw(4) << loc->allele1 << " "
    	      << setw(10) << label[p] << " "
    	      << setw(8) << Y.size() << " ";

    	  if (okay)
    	    {
    	      if ( par::return_beta )
    		OUT << setw(10) << coef[p] << " ";
    	      else
    		OUT << setw(10) << exp(coef[p]) << " ";

    	      if (par::display_ci)
    		{
    		 /* OUT << setw(8) << se << " ";

    		  if ( par::return_beta )
    		    OUT << setw(8) << coef[p] - par::ci_zt * se << " "
    			<< setw(8) << coef[p] + par::ci_zt * se << " ";
    		  else
    		    OUT << setw(8) << exp(coef[p] - par::ci_zt * se) << " "
    			<< setw(8) << exp(coef[p] + par::ci_zt * se) << " ";

    		  */

    		}

    	      OUT << setw(12) << Z << " "
    		  << setw(12) << pvalue;
    	    }
    	  else
    	    {
    	      OUT << setw(10) << "NA" << " ";

    	      if (par::display_ci)
    		OUT << setw(8) << "NA" << " "
    		    << setw(8) << "NA" << " "
    		    << setw(8) << "NA" << " ";

    	      OUT << setw(12) << "NA" << " "
    		  << setw(12) << "NA";
    	    }

    	  OUT << "\n";
    	}
        }
    }


};


#endif
